from pathlib import Path
from shutil import rmtree

from devi.cli.log import deviprint, DEVI_COLORS
from devi.utils import ask_user_input
import devi.config as config

def remove_template(template_name: str, confirm: bool):
    """Remove a template listed on $DEVI_TEMPLATES by its name
    """
    name_color = DEVI_COLORS['primary']

    # Here, devi home MUST exist, (see devi.config.init)
    templates = list(Path(config.DEVI_TEMPLATES).iterdir())

    found = next((t for t in templates if t.name == template_name), None)

    if not found:
        deviprint.err(f"template {name_color(template_name)} doesn't exist")
        exit(0)

    if not confirm:
        try:
            deviprint.info(f'template {name_color(template_name)} will be deleted')
            confirm = ask_user_input('enter to remove, ctrl+c to cancel') == ''
        except KeyboardInterrupt:
            deviprint()
            deviprint.success('canceled')
            exit(0)

    if confirm:
        rmtree(found)
        deviprint.success('removed')
        exit(0)



from sys import stderr
from typing import Callable, Optional, Any

class ANSIcodes:
    """Console colors and utilities."""
    class fg:
        default = "\u001b[39m"
        black = "\u001b[30m"
        red = "\u001b[31m"
        green = "\u001b[32m"
        yellow = "\u001b[33m"
        blue = "\u001b[34m"
        magenta = "\u001b[35m"
        cyan = "\u001b[36m"
        white = "\u001b[37m"
        rgb: Callable[[int, int, int], str] = lambda r, g, b: f"\u001b[38;2;{r};{g};{b}m"

    class bg:
        black = "\u001b[40m"
        red = "\u001b[41m"
        green = "\u001b[42m"
        yellow = "\u001b[43m"
        blue = "\u001b[44m"
        magenta = "\u001b[45m"
        cyan = "\u001b[46m"
        white = "\u001b[47m"
        rgb: Callable[[int, int, int], str] = lambda r, g, b: f"\u001b[48;2;{r};{g};{b}m"

    reset = "\u001b[0m"
    bold = "\u001b[1m"
    italic = "\u001b[3m"
    underline = "\u001b[4m"
    reverse = "\u001b[7m"

    clear = "\u001b[2J"
    clearline = "\u001b[2K"

    up = "\u001b[1A"
    down = "\u001b[1B"
    right = "\u001b[1C"
    left = "\u001b[1D"

    nextline = "\u001b[1E"
    prevline = "\u001b[1F"

    top = "\u001b[0;0H"
    goto: Callable[[int, int], str] = lambda x, y: f"\u001b[{y};{x}H"

# Colors scheme
DEVI_COLORS: dict[str, Callable[[str], str]] = {
    'default':   lambda str: ANSIcodes.fg.default + str + ANSIcodes.reset,
    'primary':   lambda str: ANSIcodes.fg.green   + str + ANSIcodes.reset,
    'secondary': lambda str: ANSIcodes.fg.blue    + str + ANSIcodes.reset,
    'error':     lambda str: ANSIcodes.fg.red     + str + ANSIcodes.reset,
    'warning':   lambda str: ANSIcodes.fg.yellow  + str + ANSIcodes.reset,
    'info':      lambda str: ANSIcodes.fg.blue    + str + ANSIcodes.reset,
    'success':   lambda str: ANSIcodes.fg.green   + str + ANSIcodes.reset,
}

class Log:
    def __init__(self):
        pass

    def __call__(self, msg: Optional[Any]="", end="\n"):
        print(msg, end=end)

    def err(self, msg: Optional[Any]="", end='\n'):
        print(f"{DEVI_COLORS['error']('[!]')} {msg}", file=stderr, end=end)

    def warn(self, msg: Optional[Any]="", end='\n'):
        print(f"{DEVI_COLORS['warning']('[!]')} {msg}", file=stderr, end=end)

    def info(self, msg: Optional[Any]="", end='\n'):
        print(f"{DEVI_COLORS['secondary']('[i]')} {msg}", file=stderr, end=end)

    def success(self, msg: Optional[Any]="", end='\n'):
        print(f"{DEVI_COLORS['success']('[-]')} {msg}", file=stderr, end=end)

deviprint = Log()



from os import rename
from sys import exit
from shutil import rmtree
from pathlib import Path
from typing import Optional
from subprocess import CalledProcessError

from devi.utils import (
    copy_dir_content,
    read_config_file,
    ask_user_input,
    edit_file,
    DeviErrorEditorNotFound,
    DeviErrorInvalidConfigFile
)
import devi.config as config
from devi.cli.log import deviprint, DEVI_COLORS

# TODO: convert this to interactive input instead of file editing
# reasons:
# - input validation: currently, we can only validate the input after file was edited
#   - we need to validate if template name already exists
# - better feeling: we are an interactive cli application
def add_template(src_path: Path, template_name: Optional[str]):
    """Main functionality for the devi's `add` command.
    """
    # Default template name is the dir name
    if not template_name:
        template_name = src_path.resolve().name

    if not src_path.exists():
        deviprint.err(f'path \'{src_path}\' does not exist')
        exit(1)

    if not src_path.is_dir():
        deviprint.err(
            f"'{src_path.resolve()}' is not a directory, that's not how templates work"
        )
        exit(1)

    template_name = ask_user_input(f'template name', default=template_name)

    template_path = Path(config.DEVI_TEMPLATES, template_name)

    # We need to handle the case where the template already exists
    if template_path.exists():
        deviprint.err(f"template '{template_name}' already exists")
        exit(1)

    # Copy to $DEVI_TEMPLATES
    copy_dir_to_template(src_path, template_name)

    template_config_path = Path(config.DEVI_TEMPLATES, template_name, config.DEVI_TEMPLATE_TOML_FILENAME)

    # The user may have a pre-defined template config, so we need to check that
    if not template_config_path.exists():
        template_config_path.touch()
        dump_default_template_config(template_config_path, template_name)

    # Edit the new config file
    editor = config.DEVI_DEFAULT_EDITOR
    try:
        deviprint.info('opening config in editor...')
        edit_file(template_config_path)
    except CalledProcessError as e:
        deviprint.err(f'{editor} exited with code {e.returncode}')
        exit_cleanup(template_path)
    except DeviErrorEditorNotFound as e:
        deviprint.warn(f"couldn't find editor: '{editor}', do you defined it properly?")

    # try to parse the newly created configfile
    try:
        read_config_file(template_config_path)
    except DeviErrorInvalidConfigFile as e:
        deviprint.err('failed to parse your config file:')
        deviprint(e)
        exit_cleanup(template_path)
    except FileNotFoundError as e:
        # This shouldn't happen, config file was created previously
        e.add_note('This may happened because the config file that previously existed was deleted')
        exit_cleanup(template_path, only_cleanup=True)
        raise

    deviprint.info(f"new template {DEVI_COLORS['primary'](template_name)} created succesfully")

def copy_dir_to_template(src_path: Path, template_name: str):
    template_path = Path(config.DEVI_TEMPLATES, template_name)
    try:
        # This function handles
        copy_dir_content(src_path, template_path)
    except (Exception, KeyboardInterrupt) as e:
        deviprint()
        deviprint.err(f"error copying recusively to '{template_path}': {e}")
        exit_cleanup(template_path)

# TODO: update this information, because currently there is no "devi config file"
def dump_default_template_config(template_file: Path, template_name: str):
    """Dumps a default config file for your new templates.

    The default content is defined in the DEVI_TEMPLATE_TOML_CONTENT constant.
    The default name for this config file should be 'default.template.devi.toml',
    and is defined in the DEVI_TEMPLATE_TOML_FILENAME constant.
    """
    conf_str = config.DEVI_TEMPLATE_TOML_CONTENT(template_name)
    try:
        with open(template_file, 'w+') as t_conf:
            t_conf.write(conf_str)
    except (IOError, OSError):
        raise

def exit_cleanup(dir_to_delete: Optional[Path] = None, only_cleanup=False):
    if dir_to_delete:
        rmtree(dir_to_delete, ignore_errors=True)
    deviprint.err('template creation aborted')
    if not only_cleanup:
        exit(1)



import sys
import subprocess
from pathlib import Path
from shutil import rmtree
from typing import Optional

from devi.utils import (
    copy_dir_content, read_config_file, DeviErrorInvalidConfigFile
)
from devi.cli.log import deviprint, DEVI_COLORS
import devi.config as config

if not config.is_windows:
    import fcntl, termios

def create_project(
    template_name: str,
    destination:   Optional[str] = None,
    project_name:  Optional[str] = None
):
    """Create a project from an existing template"""
    template_dir = Path(config.DEVI_TEMPLATES, template_name)

    # Check if the template indeed exists
    if not template_dir.exists():
        colored_name = DEVI_COLORS['primary'](f"'{template_name}'")
        deviprint.err(f"Template {colored_name} does not exist")
        exit(1)

    # Read the template.devi.toml file
    conf_data: dict = {}
    template_conf_file = template_dir / config.DEVI_TEMPLATE_TOML_FILENAME
    try:
        conf_data = read_config_file(template_conf_file)
    except FileNotFoundError as e:
        pass
    except IOError as e:
        deviprint.err(f'error reading {template_conf_file}:\n{e}')
        exit(1)
    except DeviErrorInvalidConfigFile as e:
        deviprint.err(f"couldn't parse your {template_conf_file}: \n{e}")
        deviprint.err('aborting')
        exit(1)

    # Prefer: name in flag -> name in config -> template name
    name_to_create = project_name or conf_data.get('default_name', template_name)

    # Prefer: destination in flag -> destination in config -> error
    destination_to_create = destination or conf_data.get('destination', None)
    if not destination_to_create:
        deviprint.err(
            f"{DEVI_COLORS['primary']('destination')} is defined in {DEVI_COLORS['primary']('template.devi.toml')}, use --dest"
        )
        exit(1)
    # TODO: what if is not an string? maybe is better to validate first the whole config file

    destination_path = Path(destination_to_create).expanduser()

    if not destination_path.exists():
        deviprint(f'{destination_path} Does not exist, creating')
        destination_path.mkdir(parents=True, exist_ok=True)
    if not destination_path.is_dir():
        deviprint.err(f'invalid destination: {destination_path} (not a directory)')
        exit(1)

    new_project_dir = Path(destination_path, name_to_create)
    if new_project_dir.exists():
        deviprint.err(f"couldn't create project in {DEVI_COLORS['primary'](new_project_dir.resolve().as_posix())} (already exists)")
        exit(1)

    # Copying content and then executing the oncreate command if defined
    copy_dir_content(template_dir, new_project_dir)
    deviprint.success(f"created {DEVI_COLORS['primary'](name_to_create)} at \'{new_project_dir.resolve()}\'")

    oncreate_cmd = conf_data.get('oncreate', None)
    if oncreate_cmd:
        deviprint.info(f"running {DEVI_COLORS['secondary']('oncreate')} command:")
        deviprint(f'> {oncreate_cmd}')
        # Run the command and pipe the stdout and stderr to this process
        output = subprocess.run(oncreate_cmd,
            shell=True, cwd=new_project_dir,
            stdout=sys.stdout, stderr=sys.stderr
        )
        if output.returncode != 0:
            deviprint.err(f"{DEVI_COLORS['secondary']('oncreate')} command failed with code {output.returncode}")
            deviprint.err('project was not created')
            rmtree(new_project_dir)
            exit(output.returncode)

    # Remove all the *.devi.* files/directories from the created project
    # (see README)
    for f in new_project_dir.rglob('*.devi.*'):
        if f.is_dir():
            rmtree(f)
        else:
            f.unlink()

    wanna_change_dir = conf_data.get('change_dir', False)
    if wanna_change_dir:
        if config.is_windows:
            deviprint.warn('warning: `change_dir` is not implemented yet on Windows')
            exit(0)

        dir_to_change = new_project_dir.resolve().as_posix()
        cmd = f"cd {dir_to_change}\n"
        for c in cmd:
            fcntl.ioctl(sys.stdin, termios.TIOCSTI, c.encode())



import os
import platform
from pathlib import Path
from devi.cli.log import deviprint, DEVI_COLORS
from devi.toml_parser import tomllib
from typing import Callable

is_windows = platform.system() == 'Windows'

DEVI_TEMPLATES: Path
DEVI_CONFIG: Path
DEVI_TEMPLATE_TOML_CONTENT: Callable[[str], str]
DEVI_DEFAULT_EDITOR: str
DEVI_COLORS: dict[str, Callable[[str], str]]

def init():
    """
    Initialize the config variables for devi.
    """
    global DEVI_TEMPLATES
    global DEVI_CONFIG

    global DEVI_TEMPLATE_TOML_FILENAME
    global DEVI_TEMPLATE_TOML_CONTENT
    global DEVI_DEFAULT_EDITOR

    global DEVI_COLORS # defined in log.py

    is_custom_user_dir = False
    if (devi_templates_dir := os.getenv('DEVI_TEMPLATES')):
        DEVI_TEMPLATES = Path(devi_templates_dir)
        is_custom_user_dir = True
    else:
        DEVI_TEMPLATES = Path.home() / '.templates'

    try:
        # exists ok
        if not DEVI_TEMPLATES.exists():
            DEVI_TEMPLATES.mkdir()
            deviprint.success(
                f'we just setup \'{DEVI_TEMPLATES.resolve().as_posix()}\' just for you!'
            )
            if not is_custom_user_dir:
                deviprint.info(f"if you want to override it, define {DEVI_COLORS['info']('$DEVI_TEMPLATES')}\n")
    except Exception as e:
        deviprint.err(f'couldn\'t create $DEVI_TEMPLATES at \'{DEVI_TEMPLATES}\'. exiting')
        deviprint.err(f'error received: {e}')
        exit(1)

    # Config file for devi is ~/.devi.toml
    # (TODO: put this in a better cross-platform place)
    DEVI_CONFIG = Path.home() / '.devi.toml'

    try:
        with open(DEVI_CONFIG, 'rb') as f:
            conf = tomllib.load(f)
            # loading config options
            DEVI_DEFAULT_EDITOR = conf.get('editor', get_default_text_editor())
    except IOError:
        # deviprint.warn('couldn\'t found devi config')
        DEVI_DEFAULT_EDITOR = get_default_text_editor()
        pass

    DEVI_TEMPLATE_TOML_FILENAME = 'template.devi.toml'
    # TODO: make this a customizable file
    DEVI_TEMPLATE_TOML_CONTENT = lambda template_name: (
f"""# configuration for template '{template_name}'

# this description will be shown on 'devi list'
description = ''

# this name will be used as default name for 'devi create "{template_name}"'
default_name = '{template_name}'

# destination to be used when for 'devi create "{template_name}"'
# (can be relative or absolute)
destination = '.'

# command to execute after template creation
oncreate = ''

# wheter you want to 'cd' to the newly created template or not
change_dir = true
""")

def get_default_text_editor() -> str:
    """Try to get the default text editor from the environment variables.

    It falls back for notepad on Windows and
    [sensible-editor, editor, nano, vim, vi] on *nix.
    """
    # prefer VIEWER as sensible-editor does
    for editor in [os.getenv('VIEWER'), os.getenv('EDITOR')]:
        if editor: return editor

    find_cmd = 'where' if is_windows else 'which'
    null = 'NUL' if is_windows else '/dev/null'

    for editor in ['sensible-editor', 'editor', 'nano', 'vim']:
        if os.system(f'{find_cmd} {editor} >{null} 2>&1') == 0:
            return editor

    return 'notepad' if is_windows else 'vi'



from pathlib import Path
from shutil import copy
import subprocess
import tempfile

import devi.config as config
from devi.cli.log import deviprint
from devi.toml_parser import tomllib

class DeviError(Exception):
    """Base class for exceptions in this module."""
    pass

class DeviErrorEditorNotFound(DeviError):
    """User can defined its own editor for devi"""
    pass

class DeviErrorInvalidConfigFile(DeviError):
    """Thrown when the a toml config file was failed to parse"""
    pass

def edit_file(file_path: Path):
    """Opens the file with the default text editor.

    It uses config variable 'DEVI_DEFAULT_EDITOR'
    """
    try:
        subprocess.run([config.DEVI_DEFAULT_EDITOR, file_path], check=True)
    except FileNotFoundError:
        # the editor does not exist
        raise DeviErrorEditorNotFound()

def edit_tmp_file():
    tempfile.mkstemp()

def copy_dir_content(src_path: Path, dest_path: Path):
    """Copies recursively the content of a directory to another.

    dest_path should exist, otherwise this function will create it if no parents
    are needed. Otherwise, will raise FileNotFoundError error from Path.mkdir().
    """
    if not dest_path.exists():
        dest_path.mkdir(exist_ok=True, parents=False)
    if not dest_path.is_dir():
        raise RuntimeError(f'Destination path {dest_path.resolve()} is not a directory')
    if not src_path.is_dir():
        raise RuntimeError(f'Source path {src_path.resolve()} is not a directory')
    for item in src_path.iterdir():
        if item.is_dir():
            # print(f'\ncopying dir content from {(item).absolute()} to {Path(dest_path, item.name).absolute()}')
            # Call it recursively
            copy_dir_content(item, Path(dest_path, item.name))
        else:
            # print(f"copying just a file: {item.absolute()} to {dest_path.absolute()}")
            copy(item, dest_path, follow_symlinks=True)

def ask_user_input(prompt: str, default: str = ""):
    if default:
        deviprint.info(f'{prompt}: ({default}) ', end='')
    else:
        deviprint.info(f'{prompt}: ', end='')

    response = input()

    return response if response else default

def read_config_file(config_file: Path):
    """Open and parse a toml config file

    If it fails to parse the content, throws DeviErrorInvalidConfigFile
    It raise `open` errors like IOError
    Equivalent to tomllib.load(open(config_file, 'rb'))
    """
    try:
        with open(config_file, 'rb') as conf:
            return tomllib.load(conf)
    except tomllib.TOMLDecodeError as e:
        raise DeviErrorInvalidConfigFile(e)
    except Exception:
        raise



from pathlib import Path

from devi.toml_parser import tomllib
from devi.cli.log import deviprint, ANSIcodes, DEVI_COLORS
import devi.config as config

def list_templates():
    """List all the available templates in $DEVI_TEMPLATES

    It will show the 'description' field in the template.devi.toml file
    if defined. If the 'name' field is not defined, will use the dir name.
    """
    name_color = DEVI_COLORS['primary']
    mark_color = DEVI_COLORS['secondary']

    print(DEVI_COLORS['secondary'](config.DEVI_TEMPLATES.as_posix()))

    # Here, devi home MUST exist, (see devi.config.init)
    templates = list(Path(config.DEVI_TEMPLATES).iterdir())

    if len(templates) == 0:
        deviprint(mark_color('└──'))
        deviprint()
        deviprint.info(f"is kind of empty here, try {ANSIcodes.underline}{DEVI_COLORS['primary']('devi add')}")
        exit(0)

    last_template = templates[-1].name

    # To store templates that are not directories (invalid templates)
    not_dirs = []

    # Display the directory tree (non-recursive)
    for template in templates:
        template_config_path = template / config.DEVI_TEMPLATE_TOML_FILENAME

        mark = mark_color("├──" if template.name != last_template else "└──")
        deviprint(f'{mark} {name_color(template.name)}' + ANSIcodes.reset, end='')

        if not template.is_dir():
            not_dirs.append(template.name)
        if not template_config_path.exists():
            continue

        # Exists, so try to get the description
        try:
            with open(template_config_path, 'rb') as conf_file:
                description = tomllib.load(conf_file).get('description')
                # Print the template information
                deviprint(f' - {description}' if description else '')
        except (IOError, tomllib.TOMLDecodeError) as e:
            deviprint.err(f'Error reading {template_config_path}\n{e}')
            continue

    if (how_many := len(not_dirs)) > 0:
        deviprint()
        not_dirs_str = DEVI_COLORS['warning'](str(not_dirs))
        deviprint.warn(f"watch out: {not_dirs_str} {'are not directories' if how_many > 1 else 'is not a directory'}")
        deviprint.warn('you should delete it')



try:
    # python >= 3.11
    # https://docs.python.org/3/library/tomllib.html
    import tomllib # type: ignore
except ImportError:
    # backport for python >= 3.7
    # https://realpython.com/python-toml/#load-toml-with-python
    import tomli as tomllib

tomllib = tomllib



from pathlib import Path

from devi.commands.add_template import add_template
from devi.commands.list import list_templates
from devi.commands.create import create_project
from devi.commands.remove import remove_template

def main():
    """devi: a tool for managing your project templates"""
    from devi.cli.args_parser import args

    if args['add']:
        src_path: Path = Path(args['<path>'])
        template_name: str = args['<template-name>']
        add_template(src_path, template_name)

    elif args['list']:
        list_templates()

    elif args['create']:
        template_name = args['<template-name>']
        project_name = args['--name'] or args['<project-name>']
        destination = args['<destination>']
        create_project(
            template_name,
            destination,
            project_name
        )

    elif args['rm']:
        template_name = args['<template-name>']
        confirm = bool(args['-y'])
        remove_template(template_name, confirm)



devi_bin = f"\u001b[32m{'devi'}\u001b[0m"

__doc__ = f"""{devi_bin}: a tool for managing your project templates

usage:
    {devi_bin} add <path> [<template-name>]
    {devi_bin} create <template-name> [<destination>] [--name=<project-name>]
    {devi_bin} create <template-name> as <project-name> in <destination>
    {devi_bin} list
    {devi_bin} rm     <template-name> [-y]
    {devi_bin} (-h | --help) | --version

options:
    -d, --dest=<destination>    Destination directory for the new project
    -n, --name=<project-name>   Name of the new project
    -h, --help   Shows this screen.
    --version    Shows {devi_bin} version.
"""

from docopt import docopt
from devi import __version__

args = docopt(__doc__, help=True, version=__version__)



__productname__ = 'devi'
__description__ = 'devi is a tool for managing your project templates'
__version__ = '0.1.0'
__author__ = 'the devi team'
__license__ = 'MIT'

import devi.config as config

config.init()
